@@ -54,10 +54,6 @@ class HpOptimizer:
     Examples on how to use this class are in the folder "./examples".
 
     """
-    def __init__(self):
-        self.hp = {}
-        self.model = None
-
     def build_model(self, **hp) -> object:
         """
         Method used to build the model to optimize given a set of hyper-parameters.
@@ -182,12 +178,15 @@ class HpOptimizer:
         param_gen: The parameter generator.
         X: The training data. (Union[np.ndarray, torch.Tensor, pd.Dataframe, tf.Tensor])
         y: The training labels. (Union[np.ndarray, torch.Tensor])
-        n_splits: Number of split for the kfold.
+        n_splits: Number of split for the kfold. The X_val and the y_val data will be generated by the kfold and
+                  pass to the fit method ans the score method. If n_splits is set ot <= 1, no kfold is done and the
+                  inputs X, y will be pass to the fit method ans the score method.
         nb_workers: number of used cpu. Must be a int or the string "max". Default: 1.
         save_kwargs: Saving kwargs of the parameter generator.
         verbose: True to print some stats else False.
         kwargs:
             stop_criterion: If the score once reach this criterion the optimization will stop. (float)
+            reset_gen: True to reset the parameter generator before the optimization else False. (bool) Default: True.
 
         Return
         ---------
@@ -202,18 +201,18 @@ class HpOptimizer:
         stop_criterion = kwargs.get("stop_criterion", None)
         stop_criterion_trigger = False
 
-        param_gen.reset()
-        max_itr = len(param_gen)
+        if kwargs.get("reset_gen", True):
+            param_gen.reset()
 
         if verbose:
-            progress = tqdm.tqdm(range(max_itr), unit='itr', postfix="optimisation")
+            progress = tqdm.tqdm(range(param_gen.max_itr), unit='itr', postfix="optimisation")
         else:
-            progress = range(max_itr)
+            progress = range(param_gen.max_itr)
         for i in progress:
             if not bool(param_gen) or stop_criterion_trigger:
                 break
             try:
-                workers_required = min(nb_workers, max_itr - i)
+                workers_required = min(nb_workers, param_gen.max_itr - i)
                 outputs = self._process_trial_on_X_y(param_gen, X, y, n_splits, workers_required)
 
                 stop_criterion_trigger = self._post_process_trial_(
@@ -226,6 +225,9 @@ class HpOptimizer:
 
         if verbose:
             progress.close()
+
+        param_gen.save_obj(**save_kwargs)
+        param_gen.save_history(**save_kwargs)
         param_gen.save_best_param(**save_kwargs)
         param_gen.write_optimization_to_html(show=False, **save_kwargs)
         return param_gen
@@ -321,6 +323,9 @@ class HpOptimizer:
         nb_workers: Number of used cpu. Must be a int or the string "max". Default: 1.
         save_kwargs: Saving kwargs of the parameter generator.
         verbose: True to print some stats else False.
+        kwargs:
+            stop_criterion: If the score once reach this criterion the optimization will stop. (float)
+            reset_gen: True to reset the parameter generator before the optimization else False. (bool) Default: True.
 
         Return
         ---------
@@ -335,18 +340,18 @@ class HpOptimizer:
         stop_criterion = kwargs.get("stop_criterion", None)
         stop_criterion_trigger = False
 
-        param_gen.reset()
-        max_itr = len(param_gen)
+        if kwargs.get("reset_gen", True):
+            param_gen.reset()
 
         if verbose:
-            progress = tqdm.tqdm(range(max_itr), unit='itr', postfix="optimisation")
+            progress = tqdm.tqdm(range(param_gen.max_itr), unit='itr', postfix="optimisation")
         else:
-            progress = range(max_itr)
+            progress = range(param_gen.max_itr)
         for i in progress:
             if not bool(param_gen) or stop_criterion_trigger:
                 break
             try:
-                workers_required = min(nb_workers, max_itr - i)
+                workers_required = min(nb_workers, param_gen.max_itr - i)
                 outputs = self._process_trial_on_dataset(param_gen, dataset, workers_required)
 
                 stop_criterion_trigger = self._post_process_trial_(
@@ -359,6 +364,9 @@ class HpOptimizer:
 
         if verbose:
             progress.close()
+
+        param_gen.save_obj(**save_kwargs)
+        param_gen.save_history(**save_kwargs)
         param_gen.save_best_param(**save_kwargs)
         param_gen.write_optimization_to_html(show=False, **save_kwargs)
         return param_gen
@@ -494,6 +502,9 @@ class HpOptimizer:
         -------
         The output score of the trial.
         """
+        if n_splits >= 1:
+            return self._try_params_on_X_y_without_kfold(params, X, y)
+
         kf = KFold(n_splits=n_splits, shuffle=True)
 
         mean_score = 0.0
@@ -512,6 +523,33 @@ class HpOptimizer:
 
         return mean_score
 
+    def _try_params_on_X_y_without_kfold(
+            self,
+            params: dict,
+            X, y,
+    ) -> float:
+        """
+        Try a set of hyper-parameters on the model by using the methods implemented by the user.
+
+        Parameters
+        ----------
+        params: The trial hyper-parameters.
+        X: The training input data. (Union[np.ndarray, pd.DataFrame, torch.Tensor, tf.Tensor])
+        y: The training labels. (Union[np.ndarray, torch.Tensor, tf.Tensor])
+
+        Returns
+        -------
+        The output score of the trial.
+        """
+        try:
+            model = self.build_model(**params)
+            self.fit_model_(model, X, y, **params)
+            score = self.score(model, X, y, **params)
+        except Exception as e:
+            logging.error(str(e))
+            raise e
+        return score
+
     def _take_sub_X_y(
             self,
             X, y,
