@@ -1,28 +1,33 @@
-
-import matplotlib.pyplot as plt
-import numpy as np
-import tqdm
-from typing import Dict, Union, List, Iterable, Hashable, Tuple, Callable
-import pandas as pd
-import multiprocessing
-from multiprocessing import Pool
-import time
-import os
+# ------- Base import -------- #
 import logging
-import plotly.graph_objects as go
+import os
+import time
+from typing import Dict, Union, List, Iterable, Hashable, Callable, Sequence, Optional
+
+# ------- Math -------- #
+import numpy as np
+import pandas as pd
+
+# ------- Saving -------- #
+import pickle
 import json
-import warnings
+
+# ------- Plotting -------- #
+import plotly.graph_objects as go
+import matplotlib.pyplot as plt
+
 
 # ------- App Server -------- #
 import dash
 import dash_core_components as dcc
 import dash_html_components as html
 from dash.dependencies import Input, Output
-import threading
-import plotly.io as pio
 
 
 class ParameterGenerator:
+    DEFAULT_MARKER_SIZE = 10
+    DEFAULT_MARKER_COLORSCALE = "Viridis"
+    DEFAULT_FONT_SIZE = 18
 
     class Decorators:
         @classmethod
@@ -36,7 +41,7 @@ class ParameterGenerator:
             return wrapper
 
     def __init__(self,
-                 values_dict: Dict[Union[int, str], Iterable[Union[int, float]]],
+                 values_dict: Dict[Union[int, str], Sequence[Union[int, float]]],
                  **kwargs):
         """
         Used to generate the hyper-parameter (hp) score_space, generate trial parameter for the exploration and get the best
@@ -72,11 +77,7 @@ class ParameterGenerator:
         self._values_names = list(values_dict.keys())
         self._values_dict = values_dict
 
-        for p in self._values_names:
-            self._param_name_to_type[p] = type(self._values_dict[p][0])
-            if self.check_str_in_iterable(values_dict[p]):
-                self.add_conversion_tables_param_name_to_idx(p, values_dict[p])
-                values_dict[p] = self.convert_param_to_idx(p, values_dict[p])
+        self.make_param_name_to_types(values_dict)
 
         # ------- Hp score_space -------- #
         self.xx = np.meshgrid(*[values_dict[p] for p in self._values_names])
@@ -97,19 +98,32 @@ class ParameterGenerator:
         self.app_thread = None
 
         # ------- Saving -------- #
+        # TODO: need to normalize string
         self.default_save_dir = kwargs.get("save_dir", f"parameter_generators_data/{self.__class__.__name__}/")
-        self.default_save_name = kwargs.get("save_name", '-'.join(self._values_names))
+        self.default_save_name = kwargs.get("save_name", '-'.join(self._values_names).replace(' ', '_'))
 
     def __del__(self):
-        self.close_graph_server()
+        try:
+            self.close_graph_server()
+        except AttributeError:
+            pass
+
+    def make_param_name_to_types(self, values_dict):
+        for p in self._values_names:
+            self._param_name_to_type[p] = type(self._values_dict[p][0])
+            assert self.check_homogenous_type_in_iterable(self._values_dict[p]),\
+                "All objects in the same dimension must be the same type."
+            if self.check_str_in_iterable(values_dict[p]):
+                self.add_conversion_tables_param_name_to_idx(p, values_dict[p])
+                values_dict[p] = self.convert_param_to_idx(p, values_dict[p])
 
     @property
-    def bounds_names(self) -> List[str]:
+    def hp_names(self) -> List[str]:
         return self._values_names
 
     def reset(self) -> None:
         """
-        Reset the current parameter generator.
+        Reset the current the parameter generator.
         """
         self.start_time = time.time()
         self.current_itr = 0
@@ -134,7 +148,7 @@ class ParameterGenerator:
         -------
         The elapsed time of the last itr.
         """
-        return self.elapse_time_per_iteration[self.current_itr-1]
+        return self.elapse_time_per_iteration[self.current_itr - 1]
 
     def __len__(self) -> int:
         """
@@ -275,6 +289,23 @@ class ParameterGenerator:
         """
         return any([isinstance(e, str) for e in iterable])
 
+    @staticmethod
+    def check_homogenous_type_in_iterable(iterable: Sequence) -> bool:
+        """
+        Check if all object in iterable are the same type.
+        Parameters
+        ----------
+        iterable : the iterable object.
+
+        Returns
+        ----------
+        True if all object are the same type else False.
+        """
+        if len(iterable) == 0:
+            return True
+        type0 = type(iterable[0])
+        return all([isinstance(e, type0) for e in iterable])
+
     def show_expectation(self, **kwargs) -> None:
         """
         Show the expectation of hp-score_space.
@@ -403,6 +434,10 @@ class ParameterGenerator:
         kwargs:
             title: The title of the figure. (str)
             dark_mode: True to use the dark mode else False. (bool)
+            marker_size: The size of the marker showed in the plot. (float)
+            colorscale: Color scale of the markers. The colorscales from plotly can be found at
+                        https://plotly.com/python/builtin-colorscales/. (str)
+            font_size: Font size of the plot. (int)
 
         Returns
         -------
@@ -415,8 +450,10 @@ class ParameterGenerator:
                        y=x_y_dict[self._values_names[0]]['y'],
                        mode='markers',
                        name="Trial points",
-                       marker=dict(size=5, color=x_y_dict[self._values_names[0]]['y'],
-                                   colorscale='Viridis', showscale=True),
+                       marker=dict(size=kwargs.get("marker_size", ParameterGenerator.DEFAULT_MARKER_SIZE),
+                                   color=x_y_dict[self._values_names[0]]['y'],
+                                   colorscale=kwargs.get("colorscale", ParameterGenerator.DEFAULT_MARKER_COLORSCALE),
+                                   showscale=True),
                        ),
         )
 
@@ -431,7 +468,7 @@ class ParameterGenerator:
             margin=dict(t=150, b=150, l=150, r=150),
             template="plotly_dark" if kwargs.get("dark_mode", True) else "seaborn",
             font=dict(
-                size=18,
+                size=kwargs.get("font_size", ParameterGenerator.DEFAULT_FONT_SIZE),
             )
         )
 
@@ -464,6 +501,9 @@ class ParameterGenerator:
         -------
         None
         """
+        colorscale = kwargs.get("colorscale", ParameterGenerator.DEFAULT_MARKER_COLORSCALE)
+        marker_size = kwargs.get("marker_size", ParameterGenerator.DEFAULT_MARKER_SIZE)
+
         # Add dropdown
         fig.update_layout(
             updatemenus=[
@@ -474,7 +514,10 @@ class ParameterGenerator:
                                 dict(
                                     x=[x_y_dict[p_name]['x']],
                                     y=[x_y_dict[p_name]['y']],
-                                    marker=dict(color=x_y_dict[p_name]['y'], showscale=True),
+                                    marker=dict(size=marker_size,
+                                                color=x_y_dict[p_name]['y'],
+                                                colorscale=colorscale,
+                                                showscale=True),
                                 ),
                                 {
                                     # "title": f"{p_name}",
@@ -547,6 +590,10 @@ class ParameterGenerator:
             show: True to show the figure else False. (bool)
             title: The title of the figure. (str)
             dark_mode: True to use the dark mode else False. (bool)
+            marker_size: The size of the marker showed in the plot. (float)
+            colorscale: Color scale of the markers. The colorscales from plotly can be found at
+                        https://plotly.com/python/builtin-colorscales/. (str)
+            font_size: Font size of the plot. (int)
             add_best_hp_annotation: True to activate the "Predicted best hyper-parameters:" label.
             get_best_param_kwargs: The kwargs of the method get_best_params_repr.
             save: True to save the figure else False. (bool)
@@ -563,7 +610,7 @@ class ParameterGenerator:
         self._add_dropdown_html_fig_(fig, x_y_dict, **kwargs)
         self._add_annotations_to_html_fig_(fig, x_y_dict, **kwargs)
 
-        self.save_html_fig(fig, **kwargs)
+        saved_path = self.save_html_fig(fig, **kwargs)
         if kwargs.get("show", True):
             fig.show()
 
@@ -573,7 +620,7 @@ class ParameterGenerator:
             self,
             fig: go.Figure,
             **kwargs
-    ):
+    ) -> str:
         """
 
         Parameters
@@ -586,7 +633,7 @@ class ParameterGenerator:
 
         Returns
         -------
-        None
+        The path to the saved figure.
         """
         save_dir = kwargs.get("save_dir", f"{self.default_save_dir}/html_files/")
         os.makedirs(save_dir, exist_ok=True)
@@ -595,8 +642,98 @@ class ParameterGenerator:
         if kwargs.get("save", True):
             logging.info(f"Saving html fig to {path}")
             fig.write_html(path)
+        return path
 
-    def save_best_param(self, **kwargs):
+    def get_optimization_table(self, **kwargs) -> pd.DataFrame:
+        """
+        Get the optimization table. each row is a trial on hp with it's associated score.
+
+        Parameters
+        ----------
+        kwargs:
+            None
+
+        Returns
+        -------
+        The optimization table.
+        """
+        data = {"score": [], **{p: [] for p in self._values_names}}
+
+        for i, (p_trial, p_score) in enumerate(self.history):
+            for p_name in self._values_names:
+                data[p_name].append(p_trial.get(p_name, None))
+            data["score"].append(p_score)
+
+        return pd.DataFrame(data=data)
+
+    def show_optimization(self, _parameter_name: Optional[str] = None, **kwargs):
+        if _parameter_name is None:
+            self._show_full_optimization(**kwargs)
+        else:
+            _fontsize = kwargs.get("fontsize", ParameterGenerator.DEFAULT_FONT_SIZE)
+            _title = kwargs.get('title', '')
+            _save_name = kwargs.get('save_name', f'{_parameter_name}_optimization')
+            _show = kwargs.get("show", True)
+            fig, subfig = plt.subplots(1, 1, tight_layout=True)
+            subfig = self._show_optimization_of_hp(_parameter_name, subfig, **kwargs)
+            plt.legend(fontsize=_fontsize)
+            fig_dir = kwargs.get('fig_dir', 'figures/')
+            os.makedirs(f"{fig_dir}", exist_ok=True)
+            plt.savefig(f"{fig_dir}/{_save_name}.png", dpi=300)
+            if _show:
+                plt.show()
+
+    def _show_optimization_of_hp(self, _parameter_name: str, _axe: plt.Axes, **kwargs):
+        _optimization_table = self.get_optimization_table(**kwargs)
+        # get args
+        _fontsize = kwargs.get("fontsize", ParameterGenerator.DEFAULT_FONT_SIZE)
+        _title = kwargs.get('title', '')
+
+        # plot sub fig
+        _x = _optimization_table[_parameter_name].to_numpy()
+        _scores = _optimization_table["score"].to_numpy()
+        _axe.plot(_x, _scores, marker='.', label=f"{_parameter_name}")
+        # _axe.fill_between(_x, mean_dim_f_hat, mean_dim_f_hat + mean_dim_std_hat, alpha=0.4)
+        # _axe.fill_between(_x, mean_dim_f_hat, mean_dim_f_hat - mean_dim_std_hat, alpha=0.4)
+
+        if _parameter_name in self._param_idx_to_name:
+            _axe.set_xticks(list(range(len(_x))))
+            _axe.set_xticklabels(self.convert_idx_to_param(_parameter_name, _x))
+        _axe.set_xlabel("hp space [-]", fontsize=_fontsize)
+        _axe.set_ylabel("Score [-]", fontsize=_fontsize)
+        _axe.set_title(f"{_parameter_name}", fontsize=_fontsize)
+        return _axe
+
+    def _show_full_optimization(self, **kwargs):
+        # get args
+        _fontsize = kwargs.get("fontsize", ParameterGenerator.DEFAULT_FONT_SIZE)
+        _title = kwargs.get('title', '')
+        _save_name = kwargs.get('save_name', 'full_optimization')
+        _show = kwargs.get("show", True)
+
+        # iterate on parameters
+        k = int(np.ceil(np.sqrt(len(self._values_names))))
+        j = 0
+        fig, subfigs = plt.subplots(k, k, tight_layout=True)
+        subfigs_list = subfigs.reshape(-1)
+        fig.suptitle(_title, fontsize=_fontsize)
+        for i, param_name in enumerate(self._values_names):
+            subfig = subfigs_list[i]
+            subfig = self._show_optimization_of_hp(param_name, subfig, **kwargs)
+            j = i
+
+        for i in range(j + 1, len(subfigs_list)):
+            subfig = subfigs_list[i]
+            subfig.set_axis_off()
+
+        plt.legend(fontsize=_fontsize)
+        fig_dir = kwargs.get('fig_dir', 'figures/')
+        os.makedirs(f"{fig_dir}", exist_ok=True)
+        plt.savefig(f"{fig_dir}/{_save_name}.png", dpi=300)
+        if _show:
+            plt.show()
+
+    def save_best_param(self, **kwargs) -> str:
         """
         Save the best hyper-parameters found in a json file.
 
@@ -608,20 +745,86 @@ class ParameterGenerator:
 
         Returns
         -------
-        None
+        The path to the saved file.
         """
         save_dir = kwargs.get("save_dir", f"{self.default_save_dir}/optimal_hp/")
         save_name = kwargs.get("save_name", f"{self.default_save_name}-opt_hp")
+        save_path = f'{save_dir}/{save_name}.json'
         os.makedirs(save_dir, exist_ok=True)
         # save_path = save_dir + '/' + save_name
         # np.save(save_path + ".npy", self.get_best_param(), allow_pickle=True)
 
-        with open(f'{save_dir}/{save_name}.json', 'w') as f:
+        with open(save_path, 'w') as f:
             json.dump(self.get_best_param(), f, indent=4)
+        return save_path
 
-    def save_obj(self):
-        raise NotImplementedError()
+    def save_history(self, **kwargs) -> str:
+        """
+        Save the trials history in a json file.
+
+        Parameters
+        ----------
+        kwargs:
+            save_dir: The saving directory. (str)
+            save_name: The saving name. (str)
+
+        Returns
+        -------
+        The path to the saved file.
+        """
+        save_dir = kwargs.get("save_dir", f"{self.default_save_dir}/history/")
+        save_name = kwargs.get("save_name", f"{self.default_save_name}-history")
+        save_path = f'{save_dir}/{save_name}.json'
+        os.makedirs(save_dir, exist_ok=True)
+
+        with open(save_path, 'w') as f:
+            json.dump(self.history, f, indent=4)
+        return save_path
+
+    def save_obj(self, **kwargs) -> str:
+        """
+        Save the ParameterGenerator object in a pickle file.
+
+        Parameters
+        ----------
+        kwargs:
+            save_dir: The saving directory. (str)
+            save_name: The saving name. (str)
+
+        Returns
+        -------
+        The saving path.
+        """
+        save_dir = kwargs.get("save_dir", f"{self.default_save_dir}/obj/")
+        save_name = kwargs.get("save_name", f"{self.default_save_name}-obj")
+        os.makedirs(save_dir, exist_ok=True)
+
+        path = f'{save_dir}/{save_name}.pkl'
+        with open(path, 'wb') as f:
+            pickle.dump(self, f)
+
+        return path
 
     @staticmethod
-    def load_obj(path: str):
-        raise NotImplementedError()
+    def load_obj(path: str, **kwargs) -> 'ParameterGenerator':
+        """
+        Load the ParameterGenerator and reset the start time.
+
+        Parameters
+        ----------
+        path: path to the object (returned value of ParameterGenerator.save_obj).
+        kwargs:
+            None
+
+        Returns
+        -------
+        The ParameterGenerator object.
+        """
+        with open(path, 'rb') as f:
+            obj = pickle.load(f)
+
+        if not isinstance(obj, ParameterGenerator):
+            raise TypeError(f"The object at {path} is not a ParameterGenerator")
+
+        obj.start_time = time.time()
+        return obj
